import { Observable, Subject, BehaviorSubject } from 'rxjs/Rx';
import { Http, Response } from '@angular/http';
import { Injectable, EventEmitter, Output } from '@angular/core';
import {
    INgDataGridOperation, INgDataGridData,
    IPaging, INgDataGridDataConfig,
    INgDataGridPagingConfig, INgDataGridServerConfig,
    INgDataGridChangeConfig
} from './ng-datagrid.interfaces';
import './../utils/extensions';
import { ObservableArray } from './ObservableArray';

export class NgDataGrid<T> implements INgDataGridOperation<T> {
    data: INgDataGridData<T> = <INgDataGridData<T>>{};

    constructor(private config: INgDataGridDataConfig<T>) {
        this.data = {
            searchValue: <T>{},

            totalRows: 0,
            totalFilteredRows: 0,

            items: new ObservableArray<T>(),
            itemsFiltered: [],
            itemsOnCurrentPage: [],

            isProcessing: false
        };

        this.data.items.subscribe(items => this.updateRecords());
        this.updateRecords(config.items);
    }

    getData(config: INgDataGridChangeConfig<T>): Observable<T[]> {
        this.data.searchValue = config.searchValue || this.data.searchValue;
        this.updateFilteredRecords(config);
        return this.data.items.asObservable();
    }

    clear() {
        this.data.items.array = [];
        this.data.totalRows = 0;
        this.data.totalFilteredRows = 0;
    }

    getInfo(): string {
        return `Showing ` +
            `${this.data.totalFilteredRows} of ${this.data.totalRows} entries`;
    }

    protected updateRecords(items?: T[]) {
        if (items) {
            this.data.items.array = items;
        }
        this.data.totalRows = this.data.items.array.length;

        this.updateFilteredRecords();
    }

    protected updateFilteredRecords(config: INgDataGridChangeConfig<T> = {}) {
        this.data.searchValue = config.searchValue || this.data.searchValue;
        this.data.itemsFiltered =
            this.data.items.array.filter(item => Object.same(this.data.searchValue, item));
        this.data.totalFilteredRows = this.data.itemsFiltered.length;

        this.updateCurrentItems(config);
    }

    protected updateCurrentItems(config: INgDataGridChangeConfig<T>) {
        this.data.itemsOnCurrentPage = this.data.itemsFiltered;
    }

    private search(searchValue?: T) {
        if (typeof searchValue !== 'object') {
            throw new TypeError('searchValue should type of object.');
        }

        this.updateFilteredRecords({ searchValue });
    }

}

export class NgDataGridGridPaging<T> extends NgDataGrid<T> {
    paging: IPaging = <IPaging>{};
    public get startRow(): number {
        if (this.paging.currentPageIndex === 0)
            return 0;

        return (this.paging.currentPageIndex - 1) * this.pageSize;
    }
    public get maxPageIndex(): number {
        let index = Math.ceil(this.data.totalRows / this.pageSize);
        return index;
    }
    private get currentRowStart(): number {
        return this.data.totalRows > this.pageSize
            ? (this.startRow + 1)
            : this.data.totalRows === 0 ? 0 : 1;
    }
    private get currentRowEnd(): number {
        return (this.startRow + this.pageSize) < this.data.totalRows
            ? (this.startRow + this.pageSize)
            : this.data.totalRows;
    }

    constructor(public items: T[], public pageSize: number = 10) {
        super({
            items: items
        });
    }

    getData(config: INgDataGridChangeConfig<T>): Observable<T[]> {
        return super.getData(config);
    }

    updateCurrentItems(config: INgDataGridChangeConfig<T>) {
        this.paging = <IPaging>{};
        this.paging.currentPageIndex = config.pagingNumber || 1;
        this.pageSize = config.pageSize || 10;
        this.data.itemsOnCurrentPage =
            this.data.itemsFiltered.slice(this.startRow,
                this.startRow + this.pageSize);
    }

    getInfo(): string {
        return `Showing ${this.currentRowStart} to ` +
            `${this.currentRowEnd} of ${this.data.totalRows} entries`;
    }
}

export class NgDataGridGridServer<T> extends NgDataGrid<T> {
    paging: IPaging;
    private draw: number = 0;

    constructor(public pageSize: number, private url: string,
        private showProcessing: boolean,
        private deferRendering: boolean,
        private http: Http) {
        super({});
        if (!deferRendering) {
            this.getData();
        }
    }

    getData(): Observable<T[]> {
        this.data.isProcessing = true;
        let observable = this.http.get(this.url, this.buildQuery())
            .do(_ => this.data.isProcessing = false)
            .map(response => <any>response.json())
            .map(response => response);

        observable.subscribe(item => this.data.items.array = item.data);
        return observable;
    }

    clear() {
        super.clear();
    }

    private buildQuery() {
        this.draw++;
        return {
            body: {
                draw: this.draw,
                search: this.data.searchValue,
                length: this.pageSize,
                //start: this.startRow
            }
        };
    }
}
